# Linear Algebra
#
#   by Shin-ichiro Hara
#
# Version 1.00 (2001.07.25)

require "matrix-algebra"
require "gaussian-elimination"
require "polynomial"
require "polynomial-factor"
require "algebraic-equation"

module InnerProductSpace
  def inner_product(other)
    ip = defined?(ground) ? ground.zero : first.zero
    each_with_index do |x, i|
      ip += x * other[i]
    end
    ip
  end
  
  def inner_product_complex(other)
    inner_product(other.conjugate)
  end

  def norm2
    inner_product(self)
  end
end

module Orthogonalization
  def orthogonalize
    orth_basis = []
    vectors.each_with_index do |b, i|
      orth_basis.push b
      if i > 0
	orth_basis[0...i].each do |f|
	  orth_basis[i] -= b.inner_product(f) / f.norm2 * f
#	  orth_basis[i] = f.norm2*orth_basis[i] - b.inner_product(f)*f #also
	end
      end
    end
    type.collect_column{|j| orth_basis[j]}
  end
end

class MatrixAlgebra
  include Orthogonalization

  class Vector < MatrixAlgebra
    include InnerProductSpace
  end
  class CoVector < MatrixAlgebra
    include InnerProductSpace
  end

  class SquareMatrix < MatrixAlgebra
    def diagonalize
      chp = char_polynomial(Polynomial(ground, "t"))

      facts = chp.factorize
      mdf = MinimalDecompositionField(chp, facts)
      nu = SquareMatrix(mdf, size)
      
      espaces = {}
      evalues = []
      evectors = []
      roots = []
      k = 0
      facts.each do |f, n|
	dim = f.deg
	if dim <= 0 # 0 might be occurred
	  next
	elsif dim == 1
	  evs =  [-f[0]/f[1]]
	else
	  evs = mdf.vars[k, dim].reverse
	  roots.push [f, evs]
	  k += dim
	end
	
	evs.each do |evalue|
	  ks = nu.const(evalue) - self
	  kb = ks.kernel_basis
	  espaces[evalue] = kb
	  kb.each do |v|
	    evalues.push evalue
	    evectors.push v
	  end
	end
      end

      raise "can't Diagonalize" if evalues.size < size

      s = SquareMatrix(mdf, size)
      r = s.collect_column{|i| evectors[i]}

      [mdf, roots, r, evalues, evectors, espaces, chp, facts]
    end
  end
end

if __FILE__ == $0
  require "rational"
  class Rational < Numeric
    def inspect; to_s; end
  end

  def shownorms(m)
    str = ""
    m.each_column do |v|
      v.extend InnerProductSpace
      str << "Sqrt(#{v.norm2}), "
    end
    puts str.sub(/,\s*$/, "")
  end
  
  def analize_symmetric_matrix(a)
    puts "A = "; a.display; puts

    extfield, roots, tmatrix, evalues, evectors, espaces,
      chpoly, facts = a.diagonalize
    
    puts "Charactristic Polynomial: #{chpoly}"
    puts "                          => #{facts}"
    puts

    puts "Algebraic Numbers:"
    roots.each do |po, rs|
      puts "#{rs.join(', ')} : roots of #{po} == 0"
    end
    puts

    puts "Eigen Spaces: "
    evalues.uniq.each do |ev|
      puts "W_{#{ev}} = <#{espaces[ev].join(', ')}>"
    end
    
    puts

    puts "P = "; tmatrix.display; puts
    puts "== orthogonalize ==>"
    t = tmatrix.orthogonalize
    puts "P = "; t.display; puts

    q = t.inverse
    puts "P^-1 = "; q.display; puts
    puts "P^-1 * A * P = "; (q * a * t).display; puts
    
    print "size of P's vector is: "
    shownorms(t)
    print "size of P^-1's vector is: "
    shownorms(q)
  end


  dimension = 3
  case dimension
  when 2
    Aa = [[2, 2], [2, -1]]
  when 3
#    d0, d1, d2, k0, k1, k2 = 1, 1, 1, -1, -1, -1
#    d0, d1, d2, k0, k1, k2 = 1, 0, 1, -1, -1, -1
    d0, d1, d2, k0, k1, k2 = 1, 0, 2, -1, -1, -1
    Aa = [[d0, k0, k1], [k0, d1, k2], [k1, k2, d2]]
  else
    raise
  end

  M = SquareMatrix(Rational, Aa.size)
  A = M[*Aa]
  analize_symmetric_matrix(A)
end
